@@ -78,6 +78,7 @@
 # endif
 #endif
 
+#include <stddef.h>
 #include <stdlib.h>
 #include <stdio.h>
 #include <stdarg.h>
@@ -321,9 +322,11 @@ static int
 WriteStringLen(int fd, const char *str, int length)
 {
   int type = TYPE_STRING;
+  static char zero = 0;
   if ( (Write(fd, &type, sizeof(type)) < 0) ||
        (Write(fd, &length, sizeof(length)) < 0) ||
-       (Write(fd, str, length+1) < 0) )
+       (Write(fd, str, length) < 0) ||
+       (Write(fd, &zero, 1) < 0) )
     return -1;
   return 1;
 }
@@ -1097,15 +1100,19 @@ SaveStatic(void)
 {
   SavedStatic *storage = 0;
   char *value = getenv(ENV_DJVU_STORAGE_PTR);
+  int pid = -1;
   if (value) 
-    sscanf(value, "%p", &storage);
+    sscanf(value, "%p-%d", &storage, &pid);
+  if (pid != getpid())
+    storage = 0;
   if (! storage)
     {
       char *buffer = malloc(128);
       if (buffer) {
         storage = malloc(sizeof(SavedStatic));
         if (storage) {
-          sprintf(buffer, ENV_DJVU_STORAGE_PTR "=%p", (void*)storage);
+          sprintf(buffer, ENV_DJVU_STORAGE_PTR "=%p-%d", 
+                  (void*)storage, getpid());
           putenv(buffer);
         }
       }
@@ -1130,9 +1137,12 @@ LoadStatic(void)
         nsdejavu.so */
 {
   SavedStatic *storage = 0;
+  int pid = -1;
   char *value = getenv(ENV_DJVU_STORAGE_PTR);
   if (value) 
-    sscanf(value, "%p", &storage);
+    sscanf(value, "%p-%d", &storage, &pid);
+  if (pid != getpid())
+    storage = 0;
   if (storage)
     {
       pipe_read = storage->pipe_read;
@@ -1152,8 +1162,8 @@ LoadStatic(void)
  ********************************** Callbacks **********************************
  *******************************************************************************/
 
-static int  Detach(void * id);
-static int  Resize(void * id);
+static int  Detach(void *id);
+static int  Resize(void *id);
 static void CloseConnection(void);
 static int  IsConnectionOK(int);
 static void ProgramDied(void);
@@ -1651,7 +1661,7 @@ CloseConnection(void)
 }
 
 static int
-Resize(void * id)
+Resize(void *id)
 {
   /* Instead of selecting ConfigureEvent in the application I catch
      resizeCallback here and send the appropriate request to the
@@ -1711,7 +1721,7 @@ Detach(void * id)
 }
 
 static int
-Attach(Display * displ, Window window, void * id)
+Attach(Display *displ, NPWindow *npw, void *id)
 {
   char *displ_str;
   Instance *inst;
@@ -1720,6 +1730,7 @@ Attach(Display * displ, Window window, void * id)
   Visual *visual;
   char protocol_str[128]; 
   XWindowAttributes attributes;
+  Window window = (npw) ? (Window)npw->window : 0;
   int width, height;
 #if USE_XT
   XtAppContext app_context = 0;   
@@ -1782,7 +1793,11 @@ Attach(Display * displ, Window window, void * id)
       /* Prepare protocol string */
       protocol_str[0]=0;
       if (inst->xembed_mode)
-        strcpy(protocol_str, "XEMBED");
+        {
+          width = (npw) ? npw->width : width;
+          height = (npw) ? npw->height : height;
+          strcpy(protocol_str, "XEMBED");
+        }
 #if USE_XT
       else
         {
@@ -1870,10 +1885,8 @@ StartProgram(void)
   char *ptr, *p, *q;
   struct stat st;
   int s;
-  
   if (IsConnectionOK(TRUE)) 
     return 0;
-  
   path = GetViewerPath();
   if (!path || !path[0])
     return -1;
@@ -1890,7 +1903,6 @@ StartProgram(void)
     return -1;
   rev_pipe = fd[0];
   _rev_pipe = fd[1];
-
   /* We want to wait for this child. */
   sigsave = (void*)signal(SIGCHLD,SIG_DFL);
   pid  = vfork();
@@ -1920,13 +1932,11 @@ StartProgram(void)
          we will inherit some from Netscape if we don't close them. */
       for(s=8;s<1024;s++) 
         close(s);
-      
       /* This is needed for RedHat's version of Netscape. */
       UnsetVariable("LD_PRELOAD");
       UnsetVariable("XNLSPATH");
       /* This is needed to disable session management in Qt */
       UnsetVariable("SESSION_MANAGER");      
-      
       /* Old autoinstaller fails to set the "executable" flag. */
       if (stat(path, &st)>=0)
         {
@@ -1941,12 +1951,10 @@ StartProgram(void)
       fflush(stderr);
       _exit(1);
     }
-  
   /* Parent */
   close(_pipe_write);
   close(_pipe_read);
   close(_rev_pipe);
-  
   /* Wait for the primary child */
   waitpid(pid, &s, 0);
   signal(SIGCHLD, sigsave);
@@ -1955,6 +1963,7 @@ StartProgram(void)
       CloseConnection();
       return -1;
     }
+  /* Determine scriptability */
   scriptable = 0;
   xembedable = 0;
   for (p=ptr; *p; p++)
@@ -1974,6 +1983,14 @@ StartProgram(void)
       p = q;
     }
   free(ptr);
+  /* Obtain scriptability atoms */
+  if (scriptable && !npid_version)
+    {
+      npid_getdjvuopt = NPN_GetStringIdentifier("getdjvuopt");
+      npid_setdjvuopt = NPN_GetStringIdentifier("setdjvuopt");
+      npid_onchange = NPN_GetStringIdentifier("onchange");
+      npid_version = NPN_GetStringIdentifier("version");
+    }
   return 1;
 }
 
@@ -2245,19 +2262,6 @@ NPP_Initialize(void)
   LoadStatic();
   if (pipe(delay_pipe) < 0)
     return NPERR_GENERIC_ERROR;
-  if (!IsConnectionOK(TRUE))
-    {
-      CloseConnection();
-      if (StartProgram() < 0)
-        return NPERR_GENERIC_ERROR;
-    }
-  if (scriptable)
-    {
-      npid_getdjvuopt = NPN_GetStringIdentifier("getdjvuopt");
-      npid_setdjvuopt = NPN_GetStringIdentifier("setdjvuopt");
-      npid_onchange = NPN_GetStringIdentifier("onchange");
-      npid_version = NPN_GetStringIdentifier("version");
-    }
   return NPERR_NO_ERROR;
 }
 
@@ -2300,7 +2304,6 @@ NPP_New(NPMIMEType mime, NPP np_inst, uint16 np_mode, int16 argc,
   int i;
   if (!IsConnectionOK(TRUE))
     {
-      fprintf(stderr, "nsdejavu: restarting djview (reload the page.)\n");
       CloseConnection();
       StartProgram();
     }
@@ -2436,7 +2439,7 @@ NPP_SetWindow(NPP np_inst, NPWindow * win_str)
   new_window = (win_str) ? (Window) win_str->window : 0;
   if (cur_window)
     {
-      if (new_window==cur_window)
+      if (new_window == cur_window)
         {
           Resize(id);
           return NPERR_NO_ERROR;
@@ -2454,7 +2457,7 @@ NPP_SetWindow(NPP np_inst, NPWindow * win_str)
         displ = ((NPSetWindowCallbackStruct *)(win_str->ws_info))->display;
       if (!IsConnectionOK(FALSE)) 
         return NPERR_GENERIC_ERROR;
-      if (Attach(displ, new_window, id) < 0)
+      if (Attach(displ, win_str, id) < 0)
         {
           ProgramDied();
           return NPERR_GENERIC_ERROR;
@@ -2664,8 +2667,9 @@ NPP_GetValue(NPP np_inst, NPPVariable variable, void *value)
       {
         void * id = np_inst->pdata;
         Instance *inst = map_lookup(&instance, id);
-        if (! inst || ! inst->xembed_mode) break;
+        if (! inst ) break;
         err = NPERR_NO_ERROR;
+        if (! inst->xembed_mode) break;
         *((NPBool*)value) = TRUE;
       }
       break;
