@@ -36,6 +36,7 @@
 #include <QString>
 #include <QTimer>
 #include <QUrl>
+#include <QWidget>
 
 #include "qdjvu.h"
 #include "qdjview.h"
@@ -45,6 +46,7 @@
 #include <libdjvu/miniexp.h>
 #include <libdjvu/ddjvuapi.h>
 
+#include <stddef.h>
 #include <stdlib.h>
 #include <stdio.h>
 #include <errno.h>
@@ -52,8 +54,37 @@
 #include <unistd.h>
 #include <fcntl.h>
 
-#ifdef Q_WS_X11
+#if QT_VERSION >= 0x50000
+# include <QWindow>
+# if QT_X11EXTRAS_LIB
+#  define HAVE_XCB 1
+#  include <QX11Info>
+# endif
+#endif
+
+#if QT_VERSION < 0x50000
+# if defined(Q_WS_X11)
+#  include <QX11Info>
+#  include <QX11EmbedWidget>
+# endif
+#endif
+
+#if QT_VERSION >= 0x50000
+# define HAVE_QT5EMBED 1
+#endif
+#if QT_VERSION < 0x50000 && defined(Q_WS_X11)
+# define HAVE_QX11EMBED 1
+#endif
+
+#if WITH_X11
 # ifndef X_DISPLAY_MISSING
+#  if QT_VERSION < 0x50000
+#   ifdef Q_WS_X11
+#    define HAVE_X11 1
+#   endif
+#  endif
+# endif
+# if HAVE_X11
 #  include <X11/Xlib.h>
 #  include <X11/Xutil.h>
 #  include <X11/Xatom.h>
@@ -65,18 +96,10 @@
 #  undef RevertToParent
 #  undef GrayScale
 #  undef CursorShape
-#  define HAVE_X11 1
 # endif
 #endif
 
-#ifdef Q_WS_X11
-# include <QX11Info>
-# include <QX11EmbedWidget>
-# define HAVE_QX11EMBED 1
-#endif
 
-// declared in djview.cpp
-extern void setupApplication(void);
 
 
 // ========================================
@@ -101,7 +124,10 @@ class QDjViewPlugin::Forwarder : public QObject
 public:
   QDjViewPlugin * const dispatcher;
   Forwarder(QDjViewPlugin *dispatcher);
+#if HAVE_X11
   virtual bool eventFilter(QObject*, QEvent*);
+#endif
+
 public slots:
   void containerClosed();
   void showStatus(QString message);
@@ -140,7 +166,8 @@ struct QDjViewPlugin::Instance
   QPointer<QDjViewPlugin::Document> document;
   QPointer<QWidget>                 shell;
   QPointer<QDjView>                 djview;
-  XID                               container;
+  QPointer<QObject>                 containerwin;
+  WId                               containerid;
   QStringList                    args;
   QByteArray                     saved;
   int                            savedformat;
@@ -165,7 +192,6 @@ struct QDjViewPlugin::Instance
 static Atom wm_state;
 static Atom wm_client_leader;
 
-
 static Window 
 x11ToplevelWindow(Display *dpy, Window start)
 {
@@ -201,7 +227,6 @@ x11ToplevelWindow(Display *dpy, Window start)
   return shell;
 }
 
-
 static void
 x11SetTransientForHint(QWidget *widget)
 {
@@ -248,7 +273,6 @@ x11SetTransientForHint(QWidget *widget)
     }
 }
 
-
 #endif
 
 
@@ -323,7 +347,12 @@ QDjViewPlugin::Forwarder::Forwarder(QDjViewPlugin *dispatcher)
 void 
 QDjViewPlugin::Forwarder::containerClosed()
 {
-#if HAVE_QX11EMBED
+#if HAVE_QT5EMBED
+  QDjView *shell = qobject_cast<QDjView*>(sender());
+  Instance *instance = dispatcher->findInstance(shell);
+  if (instance && dispatcher->xembedFlag)
+    instance->destroy();
+#elif HAVE_QX11EMBED
   QX11EmbedWidget *shell = qobject_cast<QX11EmbedWidget*>(sender());
   Instance *instance = dispatcher->findInstance(shell);
   if (instance && dispatcher->xembedFlag)
@@ -390,7 +419,6 @@ QDjViewPlugin::Forwarder::continueExec()
   dispatcher->continueExec();
 }
 
-
 #if HAVE_X11
 static bool x11EventFilter(void *message, long *)
 {
@@ -410,6 +438,7 @@ static bool x11EventFilter(void *message, long *)
 }
 #endif
 
+#if HAVE_X11
 bool 
 QDjViewPlugin::Forwarder::eventFilter(QObject *o, QEvent *e)
 {
@@ -418,7 +447,6 @@ QDjViewPlugin::Forwarder::eventFilter(QObject *o, QEvent *e)
       QWidget *w = static_cast<QWidget*>(o);
       switch( e->type() )
         {
-#ifdef Q_WS_X11
           // Try to fix transient windows properties.
           // - this does not work too well...
         case QEvent::Show:
@@ -429,7 +457,6 @@ QDjViewPlugin::Forwarder::eventFilter(QObject *o, QEvent *e)
           if (w->windowFlags() & Qt::Window)
             x11SetTransientForHint(w);
           break;
-#endif
         default:
           break;
         }
@@ -437,6 +464,7 @@ QDjViewPlugin::Forwarder::eventFilter(QObject *o, QEvent *e)
   // Keep processing
   return false;
 }
+#endif
 
 
 
@@ -451,6 +479,7 @@ QDjViewPlugin::Instance::~Instance()
   destroy();
   delete shell;
   delete djview;
+  delete containerwin;
   if (document)
     document->deref();
   document = 0;
@@ -459,7 +488,7 @@ QDjViewPlugin::Instance::~Instance()
 
 QDjViewPlugin::Instance::Instance(QDjViewPlugin *parent)
   : url(), dispatcher(parent), 
-    document(0), shell(0), djview(0), container(0),
+    document(0), shell(0), djview(0), containerid(0),
     onchange(0)
 {
 }
@@ -492,14 +521,20 @@ QDjViewPlugin::Instance::destroy()
     }
   if (shell)
     {
-#if HAVE_X11
+#if HAVE_QT5EMBED
+      QWindow *window = shell->windowHandle();
+      if (window)
+        window->setVisible(false);
+      if (window)
+        window->setParent(0);
+#elif HAVE_X11
       Display *dpy = QX11Info::display();
       XUnmapWindow(dpy, shell->winId());  
       XReparentWindow(dpy, shell->winId(), QX11Info::appRootWindow(), 0,0);
 #endif
       shell->close();
       dispatcher->registerForDeletion(shell);
-      container = 0;
+      containerid = 0;
       shell = 0;
     }
 }
@@ -855,7 +890,7 @@ QDjViewPlugin::cmdAttachWindow()
   Instance *instance = (Instance*) readPointer(pipeRead);
   QByteArray display = readRawString(pipeRead);
   QString protocol = readString(pipeRead);
-  XID window = (XID)readInteger(pipeRead);
+  WId window = (WId)readInteger(pipeRead);
   readInteger(pipeRead);   // colormap
   readInteger(pipeRead);   // visualid
   int width = readInteger(pipeRead);
@@ -872,27 +907,20 @@ QDjViewPlugin::cmdAttachWindow()
     {
       argc = 0;
       argv[argc++] = progname;
-#if HAVE_X11
-      Display *dpy = XOpenDisplay(display);
-      if (!dpy)
+      if (! display.isEmpty())
         {
-          fprintf(stderr,"djview dispatcher: "
-                  "cannot open display %s\n", (const char*)display);
-          throw 2;
+          argv[argc++] = "-display";
+          argv[argc++] = (const char*) display;
         }
-      argv[argc++] = "-display";
-      argv[argc++] = (const char*) display;
-#endif
       application = new QDjViewApplication(argc, const_cast<char**>(argv));
       application->setQuitOnLastWindowClosed(false);
-#if HAVE_X11
-      application->setEventFilter(x11EventFilter);
-      XCloseDisplay(dpy);
-#endif
       argc = 1;
       context = application->djvuContext();
       forwarder = new Forwarder(this);
+#if HAVE_X11
+      application->setEventFilter(x11EventFilter);
       application->installEventFilter(forwarder);
+#endif
       QObject::connect(application, SIGNAL(lastWindowClosed()), 
                        forwarder, SLOT(lastViewerClosed()));
       notifier = new QSocketNotifier(pipeRead, QSocketNotifier::Read); 
@@ -903,7 +931,7 @@ QDjViewPlugin::cmdAttachWindow()
       timer->start(5*60*1000);
       QObject::connect(timer, SIGNAL(timeout()), 
                        forwarder, SLOT(quit()));
-#if HAVE_QX11EMBED
+#if HAVE_QX11EMBED || HAVE_QT5EMBED
       if (protocol.startsWith("XEMBED"))
         xembedFlag = true;
 #endif
@@ -914,7 +942,22 @@ QDjViewPlugin::cmdAttachWindow()
   QDjView *djview = instance->djview;
   if (! shell)
     {
-#if HAVE_QX11EMBED
+#if HAVE_QT5EMBED
+      if (xembedFlag)
+        {
+          djview = new QDjView(*context, instance->viewerMode, 0);
+          djview->setAttribute(Qt::WA_NativeWindow, true);
+          shell = djview;
+          QObject::connect(shell, SIGNAL(destroyed()),
+                           forwarder, SLOT(containerClosed()) );
+          shell->setObjectName("djvu_shell");
+          shell->setGeometry(0, 0, width, height);
+          QWindow *dwindow = djview->windowHandle();
+          instance->containerwin = dwindow;
+          dwindow->setParent(QWindow::fromWinId(window));
+        }
+      else
+#elif HAVE_QX11EMBED
       if (xembedFlag)
         {
           QX11EmbedWidget *embed = new QX11EmbedWidget();
@@ -923,6 +966,13 @@ QDjViewPlugin::cmdAttachWindow()
                            forwarder, SLOT(containerClosed()) );
           shell->setObjectName("djvu_shell");
           shell->setGeometry(0, 0, width, height);
+          djview = new QDjView(*context, instance->viewerMode, shell);
+          djview->setWindowFlags(djview->windowFlags() & ~Qt::Window);
+          djview->setAttribute(Qt::WA_DeleteOnClose, false);
+#if QT_VERSION >= 0x40400
+          djview->setAttribute(Qt::WA_NativeWindow, true);
+#endif
+          instance->containerid = window;
           embed->embedInto(window);
         }
       else
@@ -931,23 +981,27 @@ QDjViewPlugin::cmdAttachWindow()
           shell = new QWidget();
           shell->setObjectName("djvu_shell");
           shell->setGeometry(0, 0, width, height);
+          djview = new QDjView(*context, instance->viewerMode, shell);
+          djview->setWindowFlags(djview->windowFlags() & ~Qt::Window);
+          djview->setAttribute(Qt::WA_DeleteOnClose, false);
+#if QT_VERSION >= 0x40400
+          djview->setAttribute(Qt::WA_NativeWindow, true);
+#endif
 #if HAVE_X11
+          instance->containerid = window;
           Display *dpy = QX11Info::display();
           XReparentWindow(dpy, shell->winId(), (Window)window, 0,0);
 #else
           qWarning("djview: unable to embed the djview window.");
 #endif
         }
-      djview = new QDjView(*context, instance->viewerMode, shell);
-      djview->setWindowFlags(djview->windowFlags() & ~Qt::Window);
-      djview->setAttribute(Qt::WA_DeleteOnClose, false);
-#if QT_VERSION >= 0x40400
-      djview->setAttribute(Qt::WA_NativeWindow, true);
-#endif
-      QLayout *layout = new QHBoxLayout(shell);
-      layout->setMargin(0);
-      layout->setSpacing(0);
-      layout->addWidget(djview);
+      if (shell && djview && shell != djview)
+        {
+          QLayout *layout = new QHBoxLayout(shell);
+          layout->setMargin(0);
+          layout->setSpacing(0);
+          layout->addWidget(djview);
+        }
       QObject::connect(djview, SIGNAL(pluginStatusMessage(QString)),
                        forwarder, SLOT(showStatus(QString)) );
       QObject::connect(djview, SIGNAL(pluginGetUrl(QUrl,QString)),
@@ -956,7 +1010,6 @@ QDjViewPlugin::cmdAttachWindow()
                        forwarder, SLOT(onChange()) );
       instance->shell = shell;
       instance->djview = djview;
-      instance->container = window;
       // apply arguments
       while (instance->args.size() > 0)
         djview->parseArgument(instance->args.takeFirst());
@@ -986,11 +1039,11 @@ QDjViewPlugin::cmdResize()
   int width = readInteger(pipeRead);
   int height = readInteger(pipeRead);
 #if HAVE_X11
-  if (instance->container)
+  if (instance->containerid)
     { 
       // the plugin lies sometimes (why?)
       int x, y; unsigned int w, h, b, d; Window r;
-      if (XGetGeometry(QX11Info::display(), (Window)(instance->container), 
+      if (XGetGeometry(QX11Info::display(), (Window)(instance->containerid), 
                        &r, &x, &y, &w, &h, &b, &d))
         { 
           width = qMin(width, (int)w);
@@ -1347,7 +1400,7 @@ QDjViewPlugin::exec()
   try 
     {
       // startup message announcing capabilities
-#if HAVE_QX11EMBED
+#if HAVE_QX11EMBED || HAVE_QT5EMBED
       const char *djviewName = "DJVIEW/4 SCRIPT XEMBED";
 #else
       const char *djviewName = "DJVIEW/4 SCRIPT";
